==> bodo-2022.7.1-cp310-cp310-manylinux2014_x86_64.manylinux_2_17_x86_64.whl: bodo/io/helpers.py <==
"""
File that contains some IO related helpers.
"""
import pyarrow as pa
from mpi4py import MPI
from numba.core import types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, models, register_model, typeof_impl, unbox
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.dict_arr_ext import dict_str_arr_type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.utils.typing import BodoError


class PyArrowTableSchemaType(types.Opaque):

    def __init__(self):
        super(PyArrowTableSchemaType, self).__init__(name=
            'PyArrowTableSchemaType')


pyarrow_table_schema_type = PyArrowTableSchemaType()
types.pyarrow_table_schema_type = pyarrow_table_schema_type
register_model(PyArrowTableSchemaType)(models.OpaqueModel)


@unbox(PyArrowTableSchemaType)
def unbox_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


@box(PyArrowTableSchemaType)
def box_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return val


@typeof_impl.register(pa.lib.Schema)
def typeof_pyarrow_table_schema(val, c):
    return pyarrow_table_schema_type


@lower_constant(PyArrowTableSchemaType)
def lower_pyarrow_table_schema(context, builder, ty, pyval):
    jymkw__isq = context.get_python_api(builder)
    return jymkw__isq.unserialize(jymkw__isq.serialize_object(pyval))


def is_nullable(typ):
    return bodo.utils.utils.is_array_typ(typ, False) and (not isinstance(
        typ, types.Array) and not isinstance(typ, bodo.DatetimeArrayType))


def pa_schema_unify_reduction(schema_a, schema_b, unused):
    return pa.unify_schemas([schema_a, schema_b])


pa_schema_unify_mpi_op = MPI.Op.Create(pa_schema_unify_reduction, commute=True)
use_nullable_int_arr = True
_pyarrow_numba_type_map = {pa.bool_(): types.bool_, pa.int8(): types.int8,
    pa.int16(): types.int16, pa.int32(): types.int32, pa.int64(): types.
    int64, pa.uint8(): types.uint8, pa.uint16(): types.uint16, pa.uint32():
    types.uint32, pa.uint64(): types.uint64, pa.float32(): types.float32,
    pa.float64(): types.float64, pa.string(): string_type, pa.large_string(
    ): string_type, pa.binary(): bytes_type, pa.date32():
    datetime_date_type, pa.date64(): types.NPDatetime('ns'), pa.null():
    string_type}


def get_arrow_timestamp_type(pa_ts_typ):
    giidi__zyzh = 'ns', 'us', 'ms', 's'
    if pa_ts_typ.unit not in giidi__zyzh:
        return types.Array(bodo.datetime64ns, 1, 'C'), False
    elif pa_ts_typ.tz is not None:
        osd__hqv = pa_ts_typ.to_pandas_dtype().tz
        ewmy__lqi = bodo.libs.pd_datetime_arr_ext.get_pytz_type_info(osd__hqv)
        return bodo.DatetimeArrayType(ewmy__lqi), True
    else:
        return types.Array(bodo.datetime64ns, 1, 'C'), True


def _get_numba_typ_from_pa_typ(pa_typ: pa.Field, is_index,
    nullable_from_metadata, category_info, str_as_dict=False):
    if isinstance(pa_typ.type, pa.ListType):
        cgc__mmm, mhe__jbjji = _get_numba_typ_from_pa_typ(pa_typ.type.
            value_field, is_index, nullable_from_metadata, category_info)
        return ArrayItemArrayType(cgc__mmm), mhe__jbjji
    if isinstance(pa_typ.type, pa.StructType):
        ojqa__tzm = []
        ffuk__jsnvf = []
        mhe__jbjji = True
        for gpop__mahf in pa_typ.flatten():
            ffuk__jsnvf.append(gpop__mahf.name.split('.')[-1])
            wzfew__xyt, ogcp__hboy = _get_numba_typ_from_pa_typ(gpop__mahf,
                is_index, nullable_from_metadata, category_info)
            ojqa__tzm.append(wzfew__xyt)
            mhe__jbjji = mhe__jbjji and ogcp__hboy
        return StructArrayType(tuple(ojqa__tzm), tuple(ffuk__jsnvf)
            ), mhe__jbjji
    if isinstance(pa_typ.type, pa.Decimal128Type):
        return DecimalArrayType(pa_typ.type.precision, pa_typ.type.scale), True
    if str_as_dict:
        if pa_typ.type != pa.string():
            raise BodoError(
                f'Read as dictionary used for non-string column {pa_typ}')
        return dict_str_arr_type, True
    if isinstance(pa_typ.type, pa.DictionaryType):
        if pa_typ.type.value_type != pa.string():
            raise BodoError(
                f'Parquet Categorical data type should be string, not {pa_typ.type.value_type}'
                )
        ogis__icks = _pyarrow_numba_type_map[pa_typ.type.index_type]
        yvs__crpm = PDCategoricalDtype(category_info[pa_typ.name], bodo.
            string_type, pa_typ.type.ordered, int_type=ogis__icks)
        return CategoricalArrayType(yvs__crpm), True
    if isinstance(pa_typ.type, pa.lib.TimestampType):
        return get_arrow_timestamp_type(pa_typ.type)
    elif pa_typ.type in _pyarrow_numba_type_map:
        rhzjg__iczkn = _pyarrow_numba_type_map[pa_typ.type]
        mhe__jbjji = True
    else:
        raise BodoError('Arrow data type {} not supported yet'.format(
            pa_typ.type))
    if rhzjg__iczkn == datetime_date_type:
        return datetime_date_array_type, mhe__jbjji
    if rhzjg__iczkn == bytes_type:
        return binary_array_type, mhe__jbjji
    cgc__mmm = (string_array_type if rhzjg__iczkn == string_type else types
        .Array(rhzjg__iczkn, 1, 'C'))
    if rhzjg__iczkn == types.bool_:
        cgc__mmm = boolean_array
    fpe__hveyl = (use_nullable_int_arr if nullable_from_metadata is None else
        nullable_from_metadata)
    if fpe__hveyl and not is_index and isinstance(rhzjg__iczkn, types.Integer
        ) and pa_typ.nullable:
        cgc__mmm = IntegerArrayType(rhzjg__iczkn)
    return cgc__mmm, mhe__jbjji

==> bodo-2022.7.3-cp310-cp310-manylinux2014_x86_64.manylinux_2_17_x86_64.whl: bodo/io/helpers.py <==
"""
File that contains some IO related helpers.
"""
import pyarrow as pa
from mpi4py import MPI
from numba.core import types
from numba.core.imputils import lower_constant
from numba.extending import NativeValue, box, models, register_model, typeof_impl, unbox
import bodo
from bodo.hiframes.datetime_date_ext import datetime_date_array_type, datetime_date_type
from bodo.hiframes.pd_categorical_ext import CategoricalArrayType, PDCategoricalDtype
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.libs.binary_arr_ext import binary_array_type, bytes_type
from bodo.libs.bool_arr_ext import boolean_array
from bodo.libs.decimal_arr_ext import DecimalArrayType
from bodo.libs.dict_arr_ext import dict_str_arr_type
from bodo.libs.int_arr_ext import IntegerArrayType
from bodo.libs.str_arr_ext import string_array_type
from bodo.libs.str_ext import string_type
from bodo.libs.struct_arr_ext import StructArrayType
from bodo.utils.typing import BodoError


class PyArrowTableSchemaType(types.Opaque):

    def __init__(self):
        super(PyArrowTableSchemaType, self).__init__(name=
            'PyArrowTableSchemaType')


pyarrow_table_schema_type = PyArrowTableSchemaType()
types.pyarrow_table_schema_type = pyarrow_table_schema_type
register_model(PyArrowTableSchemaType)(models.OpaqueModel)


@unbox(PyArrowTableSchemaType)
def unbox_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return NativeValue(val)


@box(PyArrowTableSchemaType)
def box_pyarrow_table_schema_type(typ, val, c):
    c.pyapi.incref(val)
    return val


@typeof_impl.register(pa.lib.Schema)
def typeof_pyarrow_table_schema(val, c):
    return pyarrow_table_schema_type


@lower_constant(PyArrowTableSchemaType)
def lower_pyarrow_table_schema(context, builder, ty, pyval):
    ejvz__jeazn = context.get_python_api(builder)
    return ejvz__jeazn.unserialize(ejvz__jeazn.serialize_object(pyval))


def is_nullable(typ):
    return bodo.utils.utils.is_array_typ(typ, False) and (not isinstance(
        typ, types.Array) and not isinstance(typ, bodo.DatetimeArrayType))


def pa_schema_unify_reduction(schema_a, schema_b, unused):
    return pa.unify_schemas([schema_a, schema_b])


pa_schema_unify_mpi_op = MPI.Op.Create(pa_schema_unify_reduction, commute=True)
use_nullable_int_arr = True
_pyarrow_numba_type_map = {pa.bool_(): types.bool_, pa.int8(): types.int8,
    pa.int16(): types.int16, pa.int32(): types.int32, pa.int64(): types.
    int64, pa.uint8(): types.uint8, pa.uint16(): types.uint16, pa.uint32():
    types.uint32, pa.uint64(): types.uint64, pa.float32(): types.float32,
    pa.float64(): types.float64, pa.string(): string_type, pa.large_string(
    ): string_type, pa.binary(): bytes_type, pa.date32():
    datetime_date_type, pa.date64(): types.NPDatetime('ns'), pa.null():
    string_type}


def get_arrow_timestamp_type(pa_ts_typ):
    iad__dmav = 'ns', 'us', 'ms', 's'
    if pa_ts_typ.unit not in iad__dmav:
        return types.Array(bodo.datetime64ns, 1, 'C'), False
    elif pa_ts_typ.tz is not None:
        vnjzc__tdh = pa_ts_typ.to_pandas_dtype().tz
        ymemm__vacwt = bodo.libs.pd_datetime_arr_ext.get_pytz_type_info(
            vnjzc__tdh)
        return bodo.DatetimeArrayType(ymemm__vacwt), True
    else:
        return types.Array(bodo.datetime64ns, 1, 'C'), True


def _get_numba_typ_from_pa_typ(pa_typ: pa.Field, is_index,
    nullable_from_metadata, category_info, str_as_dict=False):
    if isinstance(pa_typ.type, pa.ListType):
        yomov__nyhe, ogm__ypaj = _get_numba_typ_from_pa_typ(pa_typ.type.
            value_field, is_index, nullable_from_metadata, category_info)
        return ArrayItemArrayType(yomov__nyhe), ogm__ypaj
    if isinstance(pa_typ.type, pa.StructType):
        cnla__dehh = []
        fvs__lxw = []
        ogm__ypaj = True
        for snbh__yxm in pa_typ.flatten():
            fvs__lxw.append(snbh__yxm.name.split('.')[-1])
            lqsh__niiot, vyo__ywuaa = _get_numba_typ_from_pa_typ(snbh__yxm,
                is_index, nullable_from_metadata, category_info)
            cnla__dehh.append(lqsh__niiot)
            ogm__ypaj = ogm__ypaj and vyo__ywuaa
        return StructArrayType(tuple(cnla__dehh), tuple(fvs__lxw)), ogm__ypaj
    if isinstance(pa_typ.type, pa.Decimal128Type):
        return DecimalArrayType(pa_typ.type.precision, pa_typ.type.scale), True
    if str_as_dict:
        if pa_typ.type != pa.string():
            raise BodoError(
                f'Read as dictionary used for non-string column {pa_typ}')
        return dict_str_arr_type, True
    if isinstance(pa_typ.type, pa.DictionaryType):
        if pa_typ.type.value_type != pa.string():
            raise BodoError(
                f'Parquet Categorical data type should be string, not {pa_typ.type.value_type}'
                )
        ukpym__copy = _pyarrow_numba_type_map[pa_typ.type.index_type]
        dqq__ekg = PDCategoricalDtype(category_info[pa_typ.name], bodo.
            string_type, pa_typ.type.ordered, int_type=ukpym__copy)
        return CategoricalArrayType(dqq__ekg), True
    if isinstance(pa_typ.type, pa.lib.TimestampType):
        return get_arrow_timestamp_type(pa_typ.type)
    elif pa_typ.type in _pyarrow_numba_type_map:
        cve__uooaf = _pyarrow_numba_type_map[pa_typ.type]
        ogm__ypaj = True
    else:
        raise BodoError('Arrow data type {} not supported yet'.format(
            pa_typ.type))
    if cve__uooaf == datetime_date_type:
        return datetime_date_array_type, ogm__ypaj
    if cve__uooaf == bytes_type:
        return binary_array_type, ogm__ypaj
    yomov__nyhe = (string_array_type if cve__uooaf == string_type else
        types.Array(cve__uooaf, 1, 'C'))
    if cve__uooaf == types.bool_:
        yomov__nyhe = boolean_array
    jlien__gtnsw = (use_nullable_int_arr if nullable_from_metadata is None else
        nullable_from_metadata)
    if jlien__gtnsw and not is_index and isinstance(cve__uooaf, types.Integer
        ) and pa_typ.nullable:
        yomov__nyhe = IntegerArrayType(cve__uooaf)
    return yomov__nyhe, ogm__ypaj
